# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from statsmodels.distributions.empirical_distribution import ECDF

import pysmatch.utils as uf


def plot_scores(data, yvar, control_color="#1F77B4", test_color="#FF7F0E"):
    """
    Plots the distribution of propensity scores before matching between test and control.
    中文注释: 绘制匹配前测试组与对照组的分数分布
    """
    assert 'scores' in data.columns, \
        "Propensity scores haven't been calculated, please run predict_scores()"
    sns.kdeplot(data[data[yvar] == 0].scores, label='Control', fill=True, color=control_color)
    sns.kdeplot(data[data[yvar] == 1].scores, label='Test', fill=True, color=test_color)
    plt.legend(loc='upper right')
    plt.xlim((0, 1))
    plt.title("Propensity Scores Before Matching")
    plt.ylabel("Density")
    plt.xlabel("Scores")
    plt.show()


def compare_continuous(matcher, return_table: bool = False, plot_result: bool = True):
    """
    Plots the ECDFs for continuous features before and after matching.
    中文注释: 对连续变量在匹配前后做分布对比
    """
    test_results = []
    data = matcher.data
    matched_data = matcher.matched_data
    yvar = matcher.yvar

    for col in matched_data.columns:
        if uf.is_continuous(col, matcher.X) and col not in matcher.exclude:
            trb = matcher.test[col]
            cob = matcher.control[col]
            tra = matched_data[matched_data[yvar] == 1][col]
            coa = matched_data[matched_data[yvar] == 0][col]
            xtb, xcb = ECDF(trb), ECDF(cob)
            xta, xca = ECDF(tra), ECDF(coa)

            std_diff_med_before, std_diff_mean_before = uf.std_diff(trb, cob)
            std_diff_med_after, std_diff_mean_after = uf.std_diff(tra, coa)
            pb, truthb = uf.grouped_permutation_test(uf.chi2_distance, trb, cob)
            pa, trutha = uf.grouped_permutation_test(uf.chi2_distance, tra, coa)
            ksb = round(uf.ks_boot(trb, cob, nboots=1000), 6)
            ksa = round(uf.ks_boot(tra, coa, nboots=1000), 6)

            if plot_result:
                f, (ax1, ax2) = plt.subplots(1, 2, sharey=True, sharex=True, figsize=(12, 5))
                ax1.plot(xcb.x, xcb.y, label='Control', color=matcher.control_color)
                ax1.plot(xtb.x, xtb.y, label='Test', color=matcher.test_color)
                ax1.set_title(f'''
                    ECDF for {col} before Matching
                    KS p-value: {ksb}
                    Grouped Perm p-value: {pb}
                    Std. Median Difference: {std_diff_med_before}
                    Std. Mean Difference: {std_diff_mean_before}
                ''')
                ax2.plot(xca.x, xca.y, label='Control', color=matcher.control_color)
                ax2.plot(xta.x, xta.y, label='Test', color=matcher.test_color)
                ax2.set_title(f'''
                    ECDF for {col} after Matching
                    KS p-value: {ksa}
                    Grouped Perm p-value: {pa}
                    Std. Median Difference: {std_diff_med_after}
                    Std. Mean Difference: {std_diff_mean_after}
                ''')
                ax2.legend(loc="lower right")
                plt.xlim((0, np.percentile(xta.x, 99)))
                plt.show()

            test_results.append({
                "var": col,
                "ks_before": ksb,
                "ks_after": ksa,
                "grouped_chisqr_before": pb,
                "grouped_chisqr_after": pa,
                "std_median_diff_before": std_diff_med_before,
                "std_median_diff_after": std_diff_med_after,
                "std_mean_diff_before": std_diff_mean_before,
                "std_mean_diff_after": std_diff_mean_after
            })

    var_order = [
        "var",
        "ks_before",
        "ks_after",
        "grouped_chisqr_before",
        "grouped_chisqr_after",
        "std_median_diff_before",
        "std_median_diff_after",
        "std_mean_diff_before",
        "std_mean_diff_after"
    ]
    df_result = pd.DataFrame(test_results)[var_order] if test_results else pd.DataFrame()
    return df_result if return_table else None


def compare_categorical(matcher, return_table: bool = False, plot_result: bool = True):
    """
    Plots the proportional differences of each enumerated discrete column for test and control,
    and performs a Chi-Square Test of Independence before and after matching.
    中文注释: 对分类变量在匹配前后做卡方检验并绘制比例差异图
    """
    data = matcher.data
    matched_data = matcher.matched_data
    yvar = matcher.yvar

    def prep_plot(df: pd.DataFrame, var: str, colname: str) -> pd.DataFrame:
        t = df[df[yvar] == 1]
        c = df[df[yvar] == 0]
        dummy = [i for i in t.columns if i not in (var, "match_id", "record_id", "weight")][0]
        countt = t[[var, dummy]].groupby(var).count() / len(t)
        countc = c[[var, dummy]].groupby(var).count() / len(c)
        ret = (countt - countc).dropna()
        ret.columns = [colname]
        return ret

    title_str = '''
        Proportional Difference (test-control) for {} Before and After Matching
        Chi-Square Test for Independence p-value before | after:
        {} | {}
    '''
    test_results = []
    for col in matched_data.columns:
        if (not uf.is_continuous(col, matcher.X)) and (col not in matcher.exclude):
            dbefore = prep_plot(data, col, colname="before")
            dafter = prep_plot(matched_data, col, colname="after")
            df = dbefore.join(dafter, how="outer").fillna(0)
            test_results_i = matcher.prop_test(col)
            if test_results_i is not None:
                test_results.append(test_results_i)
                if plot_result:
                    df.plot.bar(alpha=.8)
                    plt.title(title_str.format(col, test_results_i["before"], test_results_i["after"]))
                    lim = max(.09, abs(df).max().max()) + .01
                    plt.ylim((-lim, lim))
                    plt.show()

    df_test_results = pd.DataFrame(test_results)[['var', 'before', 'after']] if test_results else pd.DataFrame()
    return df_test_results if return_table else None


import pytest
import numpy as np
import pandas as pd
from pysmatch.utils import (
    drop_static_cols, is_continuous, ks_boot, chi2_distance,
    grouped_permutation_test, std_diff
)

def test_ks_boot():
    tr = np.random.randn(30)
    co = np.random.randn(30) + 1  # 有一定分布差异
    pval = ks_boot(tr, co, nboots=50)  # 小样本就行
    assert 0 <= pval <= 1, "ks_boot pval 应该在 [0, 1] 范围内"

def test_chi2_distance():
    t = np.random.randn(50)
    c = np.random.randn(50) + 0.5
    dist = chi2_distance(t, c, bins=5)
    assert dist >= 0

def test_grouped_permutation_test():
    def my_metric(a, b):
        return np.abs(np.mean(a) - np.mean(b))

    t = np.random.randn(20)
    c = np.random.randn(20)
    p_value, truth = grouped_permutation_test(my_metric, t, c, n_samples=10)
    assert 0 <= p_value <= 1

def test_std_diff():
    a = np.array([1,2,3,4])
    b = np.array([3,4,5,6])
    med_diff, mean_diff = std_diff(a, b)
    # 只要能跑到就行, 具体断言可以更细化
    assert isinstance(med_diff, float)
    assert isinstance(mean_diff, float)


import pytest
import numpy as np
import pandas as pd
from pysmatch.Matcher import Matcher

@pytest.fixture
def sample_data():
    """
    构造一个简单的测试/对照 DataFrame
    """
    np.random.seed(42)
    size = 50
    test_df = pd.DataFrame({
        'feature1': np.random.randn(size),
        'feature2': np.random.randint(0, 5, size),
        'treatment': 1
    })
    control_df = pd.DataFrame({
        'feature1': np.random.randn(size),
        'feature2': np.random.randint(0, 5, size),
        'treatment': 0
    })
    return test_df, control_df

def test_fit_scores_optuna(sample_data):
    """
    测试 fit_scores(use_optuna=True) 分支, 覆盖 Optuna Tuner
    """
    test_df, control_df = sample_data
    matcher = Matcher(test_df, control_df, yvar='treatment')
    matcher.fit_scores(use_optuna=True, model_type='linear', n_trials=1)  # 只跑1个trial省时间
    assert len(matcher.models) == 1
    assert len(matcher.model_accuracy) == 1
    assert matcher.model_type == 'linear'

def test_fit_scores_multi_model(sample_data):
    """
    测试 fit_scores 传入 nmodels>1 的分支, 走并行逻辑
    """
    test_df, control_df = sample_data
    matcher = Matcher(test_df, control_df, yvar='treatment')
    # 对小数据集可能要减少 n_jobs 或 nmodels
    matcher.fit_scores(balance=True, model_type='linear', nmodels=2, n_jobs=1)
    assert len(matcher.models) == 2
    assert len(matcher.model_accuracy) == 2

def test_match_replacement(sample_data):
    """
    测试 match 时使用 replacement=True
    """
    test_df, control_df = sample_data
    matcher = Matcher(test_df, control_df, yvar='treatment')
    matcher.fit_scores()
    matcher.predict_scores()
    matcher.match(threshold=0.01, replacement=True)
    assert len(matcher.matched_data) > 0, "匹配结果不应为空"
    # 验证 replacement 一般会产生更多匹配记录

def test_tune_threshold(sample_data):
    """
    测试 tune_threshold 方法
    """
    test_df, control_df = sample_data
    matcher = Matcher(test_df, control_df, yvar='treatment')
    matcher.fit_scores()
    matcher.predict_scores()
    # 这里 range 可稍微扩大些，当然要考虑运行时间
    rng = np.arange(0, 0.005, 0.001)
    matcher.tune_threshold(method='min', nmatches=1, rng=rng)
    # 单纯保证不报错即可；实际可细化断言

def test_record_frequency_and_weights(sample_data):
    """
    测试 record_frequency 和 assign_weight_vector
    """
    test_df, control_df = sample_data
    matcher = Matcher(test_df, control_df, yvar='treatment')
    matcher.fit_scores()
    matcher.predict_scores()
    matcher.match(threshold=0.01)
    freq_df = matcher.record_frequency()
    assert isinstance(freq_df, pd.DataFrame)
    matcher.assign_weight_vector()
    # assign_weight_vector 后 matched_data 应多出 weight 列
    assert 'weight' in matcher.matched_data.columns

def test_prop_test(sample_data):
    """
    测试 prop_test 对某个离散变量做卡方检验
    """
    test_df, control_df = sample_data
    # 人工加入个离散变量
    test_df['cat_var'] = np.random.choice(['A', 'B'], size=len(test_df))
    control_df['cat_var'] = np.random.choice(['A', 'B'], size=len(control_df))
    matcher = Matcher(test_df, control_df, yvar='treatment')
    matcher.fit_scores()
    matcher.predict_scores()
    matcher.match(threshold=0.01)
    res = matcher.prop_test('cat_var')
    # 结果应该包含 before/after p值
    assert res is not None
    assert 'before' in res and 'after' in res

def test_compare_continuous(sample_data):
    """
    测试 compare_continuous 分支
    """
    test_df, control_df = sample_data
    matcher = Matcher(test_df, control_df, yvar='treatment')
    matcher.fit_scores()
    matcher.predict_scores()
    matcher.match(threshold=0.01)
    # 注意: matched_data 不为空时，才能compare
    # 如果想避免弹出画图窗口, 可以 set plot_result=False
    df_result = matcher.compare_continuous(return_table=True, plot_result=False)
    # 检查结果是否为 DataFrame
    assert df_result is not None

def test_compare_categorical(sample_data):
    """
    测试 compare_categorical 分支
    """
    test_df, control_df = sample_data
    # 手动加个分类变量
    test_df['cat_var'] = np.random.choice(['X', 'Y'], size=len(test_df))
    control_df['cat_var'] = np.random.choice(['X', 'Y'], size=len(control_df))
    matcher = Matcher(test_df, control_df, yvar='treatment')
    matcher.fit_scores()
    matcher.predict_scores()
    matcher.match(threshold=0.01)
    df_cat_res = matcher.compare_categorical(return_table=True, plot_result=False)
    assert df_cat_res is not None
    # 结果里应包含 'var', 'before', 'after'
    if not df_cat_res.empty:
        assert all(col in df_cat_res.columns for col in ['var','before','after'])


# -*- coding: utf-8 -*-
from __future__ import print_function
import logging
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from typing import List, Optional

import pysmatch.utils as uf
from pysmatch.modeling import fit_model, optuna_tuner


logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(levelname)s - %(message)s',
    datefmt='%Y-%m-%d %H:%M:%S'
)


class Matcher:
    """
    Matcher Class -- Match data for an observational study.
    中文注释: 匹配器类, 用于观察性研究中进行配对
    """

    def __init__(self, test: pd.DataFrame, control: pd.DataFrame, yvar: str,
                 formula: Optional[str] = None, exclude: Optional[List[str]] = None):
        if exclude is None:
            exclude = []
        plt.rcParams["figure.figsize"] = (10, 5)
        aux_match = ['scores', 'match_id', 'weight', 'record_id']
        t = test.copy().reset_index(drop=True)
        c = control.copy().reset_index(drop=True)
        t = t.dropna(axis=1, how="all")
        c = c.dropna(axis=1, how="all")
        c.index += len(t)
        self.data = pd.concat([t, c], ignore_index=True)
        self.control_color = "#1F77B4"
        self.test_color = "#FF7F0E"
        self.yvar = yvar
        self.exclude = exclude + [self.yvar] + aux_match
        self.formula = formula
        self.nmodels = 1
        self.models = []
        self.swdata = None
        self.model_accuracy = []
        self.errors = 0
        self.data[yvar] = self.data[yvar].astype(int)  # should be binary 0, 1
        self.xvars = [i for i in self.data.columns if i not in self.exclude]
        self.original_xvars = self.xvars.copy()
        self.data = self.data.dropna(subset=self.xvars)
        self.matched_data = []
        self.xvars_escaped = [f"Q('{x}')" for x in self.xvars]
        self.yvar_escaped = f"Q('{self.yvar}')"
        self.X = self.data[self.xvars]
        self.y = self.data[self.yvar]
        self.test = self.data[self.data[yvar] == 1]
        self.control = self.data[self.data[yvar] == 0]
        self.testn = len(self.test)
        self.controln = len(self.control)
        if self.testn <= self.controln:
            self.minority, self.majority = 1, 0
        else:
            self.minority, self.majority = 0, 1
        logging.info(f'Formula: {yvar} ~ {"+".join(self.xvars)}')
        logging.info(f'n majority: {len(self.data[self.data[yvar] == self.majority])}')
        logging.info(f'n minority: {len(self.data[self.data[yvar] == self.minority])}')


    def fit_model(self, index: int, X: pd.DataFrame, y: pd.Series, model_type: str,
                  balance: bool, max_iter: int = 100) -> dict:
        """
        Internal helper that calls modeling.fit_model
        """
        return fit_model(index, X, y, model_type, balance, max_iter=max_iter)


    def fit_scores(self, balance: bool = True, nmodels: Optional[int] = None,
                   n_jobs: int = 1, model_type: str = 'linear',
                   max_iter: int = 100, use_optuna: bool = False,
                   n_trials: int = 10):
        """
        Fits one or multiple models to get propensity scores.
        中文注释: 训练(多个)模型以获取倾向分数
        新增: 如果 use_optuna=True, 则使用 optuna 搜索超参(只训练1个最佳模型)
        """
        import multiprocessing as mp
        from multiprocessing.pool import ThreadPool as Pool

        self.models, self.model_accuracy = [], []
        self.model_type = model_type
        num_cores = mp.cpu_count()
        logging.info(f"This computer has: {num_cores} cores, The workers will be: {min(num_cores, n_jobs)}")

        if use_optuna:
            result = optuna_tuner(self.X, self.y, model_type=model_type,
                                  n_trials=n_trials, balance=balance)
            self.models.append(result['model'])
            self.model_accuracy.append(result['accuracy'])
            logging.info(f"Optuna Best Accuracy: {result['accuracy']:.2%}")
            return

        if balance and nmodels is None:
            minor, major = [self.data[self.data[self.yvar] == i] for i in (self.minority, self.majority)]
            import numpy as np
            nmodels = int(np.ceil((len(major) / len(minor)) / 10) * 10)
        if nmodels is None:
            nmodels = 1

        nmodels = max(1, nmodels)

        if balance and nmodels > 1:
            with Pool(min(num_cores, n_jobs)) as pool:
                tasks = [
                    (i, self.X, self.y, model_type, balance, max_iter)
                    for i in range(nmodels)
                ]
                results = pool.starmap(self.fit_model, tasks)
            for res in results:
                self.models.append(res['model'])
                self.model_accuracy.append(res['accuracy'])
            import numpy as np
            logging.info(f"Average Accuracy: {np.mean(self.model_accuracy):.2%}")
        else:
            result = self.fit_model(0, self.X, self.y, model_type, balance, max_iter)
            self.models.append(result['model'])
            self.model_accuracy.append(result['accuracy'])
            logging.info(f"Accuracy: {self.model_accuracy[0]*100:.2f}%")

    def predict_scores(self):
        """
        Predict propensity scores using the trained models.
        中文注释: 使用已训练的模型预测倾向分数
        """
        if not self.models:
            logging.warning("No trained models found. Call fit_scores() first.")
            return
        model_preds = []
        for m in self.models:
            preds = m.predict_proba(self.X)[:, 1]
            model_preds.append(preds)
        model_preds = np.array(model_preds)
        scores = np.mean(model_preds, axis=0)
        self.data['scores'] = scores

    def match(self, threshold: float = 0.001, nmatches: int = 1, method: str = 'min',
              replacement: bool = False):
        """
        Finds suitable match(es) for each record in the minority dataset.
        中文注释: 调用匹配方法，对少数类样本找到匹配记录
        """
        from pysmatch.matching import perform_match
        matched_df = perform_match(
            self.data, self.yvar, threshold=threshold,
            nmatches=nmatches, method=method, replacement=replacement
        )
        self.matched_data = matched_df

    def plot_scores(self):
        """
        Plots the distribution of propensity scores before matching.
        中文注释: 绘制匹配前的倾向分数分布
        """
        from pysmatch.visualization import plot_scores
        plot_scores(self.data, self.yvar, control_color=self.control_color, test_color=self.test_color)

    def tune_threshold(self, method: str, nmatches: int = 1,
                       rng: np.ndarray = np.arange(0, .001, .0001)):
        """
        Matches data over a grid to optimize threshold value and plots results.
        中文注释: 在一系列阈值范围上做匹配并绘制保留率
        """
        from pysmatch.matching import tune_threshold
        thresholds, retained = tune_threshold(self.data, self.yvar,
                                              method=method, nmatches=nmatches, rng=rng)
        plt.plot(thresholds, retained)
        plt.title("Proportion of Data retained for grid of threshold values")
        plt.ylabel("Proportion Retained")
        plt.xlabel("Threshold")
        plt.xticks(thresholds, rotation=90)
        plt.show()

    def record_frequency(self) -> pd.DataFrame:
        """
        Calculates the frequency of specific records in the matched dataset.
        中文注释: 计算匹配后数据中记录出现的次数
        """
        if len(self.matched_data) == 0:
            logging.info("No matched data found. Please run match() first.")
            return pd.DataFrame()
        freqs = self.matched_data['match_id'].value_counts().reset_index()
        freqs.columns = ['freq', 'n_records']
        return freqs

    def assign_weight_vector(self):
        """
        Assigns an inverse frequency weight to each record in the matched dataset.
        中文注释: 为匹配后的每条记录分配一个“1 / 匹配次数”的权重
        """
        if len(self.matched_data) == 0:
            logging.info("No matched data found. Please run match() first.")
            return
        record_freqs = self.matched_data.groupby('record_id').size().reset_index(name='count')
        record_freqs['weight'] = 1 / record_freqs['count']
        self.matched_data = self.matched_data.merge(record_freqs[['record_id', 'weight']], on='record_id')

    def prop_test(self, col: str) -> Optional[dict]:
        """
        Performs a Chi-Square test of independence on <col>
        中文注释: 对某个离散变量做卡方检验
        """
        from scipy import stats
        if not uf.is_continuous(col, self.X) and col not in self.exclude:
            pval_before = round(stats.chi2_contingency(self.prep_prop_test(self.data, col))[1], 6)
            pval_after = round(stats.chi2_contingency(self.prep_prop_test(self.matched_data, col))[1], 6)
            return {'var': col, 'before': pval_before, 'after': pval_after}
        else:
            logging.info(f"{col} is a continuous variable or excluded.")
            return None

    def prep_prop_test(self, data: pd.DataFrame, var: str):
        """
        Helper method for running chi-square contingency tests.
        中文注释: 卡方检验辅助方法，补全空的类别计数
        """
        counts = data.groupby([var, self.yvar]).size().unstack(fill_value=0)
        if 0 not in counts.columns:
            counts[0] = 0
        if 1 not in counts.columns:
            counts[1] = 0
        counts = counts[[0, 1]]
        return counts.values.tolist()

    def compare_continuous(self, save: bool = False, return_table: bool = False, plot_result: bool = True):
        """
        Wrapper to call visualization.compare_continuous
        """
        from pysmatch.visualization import compare_continuous
        return compare_continuous(self, return_table=return_table, plot_result=plot_result)

    def compare_categorical(self, return_table: bool = False, plot_result: bool = True):
        """
        Wrapper to call visualization.compare_categorical
        """
        from pysmatch.visualization import compare_categorical
        return compare_categorical(self, return_table=return_table, plot_result=plot_result)